import numpy as np
import tensorflow as tf
from tensorflow.keras import Model
from tensorflow.keras.layers import BatchNormalization, Conv3D, Conv3DTranspose, Layer


# 自定义两个函数
@tf.custom_gradient
def tensor_div(a, x1):
    m = tf.math.divide_no_nan(a, x1 + 0.001)  # 0.9*x1+0.1*a)

    # m=tf.where(x1<0.01,x=tf.ones_like(a),y=tf.div_no_nan(a, x1))
    def grad(dy):
        return dy, -tf.square(
            m) * dy  # tf.where(x1<0.01, x=dy,y=-tf.square(m)*dy)#(tf.where(m>3.0, x=-9*tf.ones_like(m),y=-tf.square(m)))

    return m, grad


@tf.custom_gradient
def tensor_mul(a, x):
    m = tf.math.multiply(a, x)

    def grad(dy):
        return dy, dy * (a)

    return m, grad


# @tf.custom_gradient
def s_sigmoid(x):
    return tf.nn.softplus(x)  # tf.nn.sigmoid(x/2)*4#tf.nn.softplus(x)


# @tf.custom_gradient
def s_sigmoid1(x):
    return tf.nn.softplus(x)  # tf.nn.sigmoid(x/2)*4#sigmoid(x/2)*4


class Forward_Pass(Layer):

    def tf_fspecial_gauss(self, size, sigma1):
        x_data, y_data, z_data = np.mgrid[-size[0] // 2 + 1:size[0] // 2 + 1, -size[1] // 2 + 1:size[1] // 2 + 1,
                                 -size[2] // 2 + 1:size[2] // 2 + 1]
        #        print(x_data.shape)
        x_data = np.expand_dims(x_data, axis=-1)
        x_data = np.expand_dims(x_data, axis=-1)

        y_data = np.expand_dims(y_data, axis=-1)
        y_data = np.expand_dims(y_data, axis=-1)

        z_data = np.expand_dims(z_data, axis=-1)
        z_data = np.expand_dims(z_data, axis=-1)

        x = tf.constant(x_data, dtype=tf.float32)
        y = tf.constant(y_data, dtype=tf.float32)
        z = tf.constant(z_data, dtype=tf.float32)

        g = tf.exp(-((x ** 2 + y ** 2 + z ** 2) / (2.0 * sigma1 ** 2)))  # * tf.exp(-((z**2)/(2.0*9*sigma1**2)))
        g = g / tf.reduce_max(g)
        return g  # g/tf.reduce_max(g) #last_sum

    def gaussian_ker(self, kernel_shape, in_channels, out_channels=4, name="forward_pass", stddev=2):
        if out_channels != 4:
            raise ("Not Implemented")

        #        kernal1=self.tf_fspecial_gauss(shape,stddev*0.2)
        kernal2 = self.tf_fspecial_gauss(kernel_shape, stddev * 0.5)
        kernal3 = self.tf_fspecial_gauss(kernel_shape, stddev)
        kernal4 = self.tf_fspecial_gauss(kernel_shape, stddev * 1.5)
        kernal1 = self.tf_fspecial_gauss(kernel_shape, stddev * 2)
        rad = tf.random_uniform([1, 1, 1, in_channels, 1], minval=0.7, maxval=1, dtype=tf.float32)
        init = tf.concat([kernal1, kernal2, kernal3, kernal4], -1)
        #        print(init.shape)
        init = tf.tile(init, [1, 1, 1, in_channels, 1]) * rad
        w = tf.Variable(initial_value=init, name=name)
        return w

    def __init__(self, kernel_shape, stride, name=None, out_channels=4, ):
        super().__init__(name=name)
        self.stride = stride
        self.kernel_shape = kernel_shape
        self.out_channels = out_channels

    def build(self, input_shape):
        in_channels = input_shape[-1]
        self.w = self.gaussian_ker(kernel_shape=self.kernel_shape, in_channels=in_channels,
                                   out_channels=self.out_channels, name="conv_kernel")

    def call(self, x):
        return tf.nn.conv3d(input=x, filters=self.w, strides=self.stride, padding='SAME')


class Forward1(Layer):
    def __init__(self, name="Forward_Est", **kwargs):
        super().__init__(name=name, **kwargs)
        self.f1_1 = Forward_Pass(kernel_shape=(3, 3, 3), stride=[1, 1, 1, 1, 1], name="f1.1")
        self.f1_1_norm = BatchNormalization()
        self.f1_2 = Forward_Pass((3, 3, 3), stride=[1, 1, 1, 1, 1], name="f1.2")
        self.f1_2_norm = BatchNormalization()
        self.f1_3 = Forward_Pass((3, 3, 3), stride=[1, 1, 1, 1, 1], name="f1.3")
        self.f1_3_norm = BatchNormalization()

    def build(self, __input_shape):
        super().build(__input_shape)

    def call(self, x):
        normed_batch_t_down = x
        normed_batch_t_down_m = s_sigmoid(
            tf.tile(input=normed_batch_t_down, multiples=[1, 1, 1, 1, 4]))  # polled and activated input

        # concolution f1.1
        result_conv_f1_1 = self.f1_1(normed_batch_t_down)  # +normed_batch_t_down_m
        result_f1_1 = s_sigmoid1(self.f1_1_norm(result_conv_f1_1))

        # convolution f1.2
        result_conv_9 = self.f1_2(result_f1_1)  # +normed_batch_t_down_m
        result_f1_2 = s_sigmoid1(self.f1_2_norm(result_conv_9))  # +normed_batch_t_down_m

        dense_input = tf.concat([result_f1_1, result_f1_2], -1)  # dense connection between h1.1 and h1.2

        # convolution f1.3
        result_conv_f1_3 = self.f1_3(dense_input)  # +normed_batch_t_down_m
        normed_batch_f1_3 = self.f1_3_norm(result_conv_f1_3)
        result_f1_3 = s_sigmoid(
            normed_batch_f1_3) + normed_batch_t_down_m  # adds activated input to output of the convolution

        result = tf.reduce_mean(result_f1_3, axis=4, keepdims=True)  # average over channels
        return result


class Backward1(Layer):
    def __init__(self):
        super().__init__()
        self.b1_1 = Conv3D(8, (3, 3, 3), strides=(1, 1, 1), padding="same", name="b1.1")
        self.b1_1_norm = BatchNormalization()
        self.b1_2 = Conv3D(8, (3, 3, 3), strides=(1, 1, 1), padding="same", name="b1.2")
        self.b1_2_norm = BatchNormalization()
        self.b1_3 = Conv3D(8, (3, 3, 3), strides=(1, 1, 1), padding="same", name="b1.3")
        self.b1_3_norm = BatchNormalization()
        self.b1_4 = Conv3DTranspose(4, (2, 2, 2), strides=(2, 2, 2), padding="valid", name="b1.4")
        self.b1_4_norm = BatchNormalization()
        self.b1_5 = Conv3D(4, (3, 3, 3), strides=(1, 1, 1), padding="same", name="b1.5")
        self.b1_5_norm = BatchNormalization()

    def build(self, __input_shape):
        super().build(__input_shape)

    def call(self, x):
        result_conv_b1_1 = self.b1_1(x)
        result_b1_1 = s_sigmoid1(self.b1_1_norm(result_conv_b1_1))

        # convolution b1.2
        result_conv_b1_2 = self.b1_2(result_b1_1)
        result_b1_2 = s_sigmoid1(self.b1_2_norm(result_conv_b1_2))

        dense_input = tf.concat([result_b1_2, result_b1_1], -1)  # dense layer between b1.1 and b1.2

        # dense convolution b1.3
        result_conv_b1_3 = self.b1_3(dense_input)
        # result_conv_b1_3=result_conv_b1_3+tf.ones_like(result_conv_b1_3)
        result_b1_3 = s_sigmoid1(self.b1_3_norm(result_conv_b1_3))

        # convolution b1.4
        result_conv_b1_4 = self.b1_4(result_b1_3)
        result_b1_4 = s_sigmoid1(self.b1_4_norm(result_conv_b1_4))

        # convolution b1.5
        result_conv_b1_5 = self.b1_5(result_b1_4)
        result_b1_5 = s_sigmoid(self.b1_5_norm(result_conv_b1_5))

        return tf.reduce_mean(result_b1_5, axis=4, keepdims=True)  # reduce channels to 1


class Estimate(Layer):
    def __init__(self, name):
        super().__init__(name=name)
        self.forward1 = Forward1()
        self.h1_norm1 = BatchNormalization()
        self.backward1 = Backward1()
        self.h1_norm2 = BatchNormalization()

    def build(self, __input_shape):
        super().build(__input_shape)

    def call(self, __inputs):
        original_image = __inputs[0]
        previous_iteration = __inputs[1]

        original_image_down_scaled = tf.nn.avg_pool2d(original_image, [1, 2, 2, 1], strides=[1, 2, 2, 1],
                                                      padding='VALID', data_format='NHWC')  # pooled input
        x = self.forward1(original_image_down_scaled)
        x = tensor_div(original_image_down_scaled, x)  # tensor division as in RLD
        x = self.h1_norm1(x)
        x = self.backward1(x)

        return tensor_mul(previous_iteration, x)  # multiplication as in RLD take input as previous iteration


class Update(Layer):
    def __init__(self, name):
        super().__init__(name=name)
        self.f2_1 = Forward_Pass((4, 4, 4), stride=[1, 1, 1, 1, 1], name="f2.1")
        self.f2_1_norm = BatchNormalization()
        self.f2_2 = Forward_Pass((4, 4, 4), stride=[1, 1, 1, 1, 1], name="f2.2")
        self.f2_2_norm = BatchNormalization()

        self.DIV_NORM = BatchNormalization()

        self.b2_1 = Conv3D(8, (3, 3, 3), strides=(1, 1, 1), padding="same", name="b2.1")
        self.b2_1_norm = BatchNormalization()
        self.b2_2 = Conv3D(8, (3, 3, 3), strides=(1, 1, 1), padding="same", name="b2.2")
        self.b2_2_norm = BatchNormalization()

    def build(self, __input_shape):
        super().build(__input_shape)

    def call(self, __inputs):
        original_image = __inputs[0]
        previous_iteration = __inputs[1]
        # f2.1
        result_conv_1 = self.f2_1(original_image)
        result_conv_1_b = s_sigmoid1(self.f2_1_norm(result_conv_1))

        # f2.2
        result_conv_1_1 = self.f2_2(result_conv_1_b)  # +normed_batch_m
        normed_batch_1 = self.f2_2_norm(result_conv_1_1)
        result_prelu_1 = s_sigmoid(normed_batch_1) + s_sigmoid(
            tf.tile(input=original_image, multiples=[1, 1, 1, 1, 4]))  # add original image

        ave_1 = tf.reduce_mean(result_prelu_1, axis=4, keepdims=True)  # merge channels to 1

        EST = tensor_div(original_image, ave_1)  # RLD division
        EST = self.DIV_NORM(EST)

        # b2.1
        result_conv_2 = self.b2_1(EST)
        result_conv_2_b = s_sigmoid1(self.b2_1_norm(result_conv_2))

        # b2.2
        result_conv_2_1 = self.b2_2(result_conv_2_b)
        normed_batch_2 = self.b2_2_norm(result_conv_2_1)
        act_2 = s_sigmoid(normed_batch_2) + tf.ones_like(normed_batch_2)

        ave_2 = tf.reduce_mean(act_2, axis=4, keepdims=True)  # remove channel info
        return tensor_mul(previous_iteration,
                          ave_2)  # multiplication as in RLD take as previous iteration the output of 1st iteration


class RLN_model(Model):
    def __init__(self, name):
        super().__init__(name=name)
        self.est_1 = Estimate(name="Estimate_1")
        self.est_1_norm = BatchNormalization()
        self.update_1 = Update(name="update_1")
        self.update_1_norm = BatchNormalization()

        self.u1 = Conv3D(8, (3, 3, 3), (1, 1, 1), padding="same", name="u1")
        self.u2 = Conv3D(8, (3, 3, 3), (1, 1, 1), padding="same", name="u2")
        self.u3 = Conv3D(8, (3, 3, 3), (1, 1, 1), padding="same", name="u3")

        self.final_norm = BatchNormalization()

    def call(self, inputs):
        original_image = inputs
        estimatation1 = self.est_1((original_image, original_image))
        estimation1_normed = s_sigmoid(self.est_1_norm(estimatation1))  # result_prelu_12_1

        update1 = self.update_1((original_image, estimation1_normed))
        update1_normed = s_sigmoid(self.update_1_norm(update1))  # result_prelu_2_1

        result_conv_2_fine1 = self.u1(update1_normed)

        result_conv_2_fined = s_sigmoid1(result_conv_2_fine1)

        # here we create a tensor that has 10 channels:
        # 1-8: are the results of U1
        # 9: is the result of the h2 line
        # 10: is the result of the h1 line
        result_conv_2_fine1_c = tf.concat([result_conv_2_fined, update1_normed, estimation1_normed],
                                          -1)  # if not ok, try result_prelu_12_1

        # do a convolution that reduces channels by 2 called U2
        result_conv_2_fine = self.u2(result_conv_2_fine1_c)
        act_2_fine = s_sigmoid1(result_conv_2_fine)

        # this creates an input tensor that has:
        # 1-8 results of U1
        # 9-18: the results of U2
        Merge = tf.concat([result_conv_2_fined, act_2_fine], -1)

        # final convolution that reduces the output from 16 channels to 8 channels U3
        result_conv_13 = self.u3(Merge)  # +result_conv_2_fined
        normed_batch_13 = self.final_norm(result_conv_13)
        result_prelu_13 = normed_batch_13  # s_sigmoid(normed_batch_13)  # tf.nn.leaky_relu(normed_batch_13,name='relu_1'）

        # averge over the remaining channels as the output

        prediction = tf.reduce_mean(result_prelu_13, axis=4,
                                    keepdims=True)  # /tf.reduce_max(result_prelu_14)#*tf.reduce_max(normed_batch) #tf.multiply(result_prelu_7,result_prelu_1)

        return (prediction, estimatation1, update1)


class RLN_model_simple(Model):
    def __init__(self, name):
        super().__init__(name=name)
        self.est_1 = Estimate(name="Estimate_1")
        self.est_1_norm = BatchNormalization()
        self.update_1 = Update(name="update_1")
        self.update_1_norm = BatchNormalization()

        self.u2 = Conv3D(8, (3, 3, 3), (1, 1, 1), padding="same", name="u2")
        self.u3 = Conv3D(8, (3, 3, 3), (1, 1,1 ), padding="same", name="u3")

        self.final_norm = BatchNormalization()

    def call(self, inputs):
        original_image = inputs
        estimatation1 = self.est_1((original_image, original_image))
        estimation1_normed = s_sigmoid(self.est_1_norm(estimatation1))  # result_prelu_12_1

        # here we create a tensor that has 10 channels:
        # 1-8: are the results of U1
        # 9: is the result of the h2 line
        # 10: is the result of the h1 line
        result_conv_2_fine1_c = tf.tile(estimation1_normed, [1, 1, 1, 1, 10])
        # do a convolution that reduces channels by 2 called U2
        result_conv_2_fine = self.u2(result_conv_2_fine1_c)
        act_2_fine = s_sigmoid1(result_conv_2_fine)

        # this creates an input tensor that has:
        # 1-8 results of U1
        # 9-18: the results of U2

        # final convolution that reduces the output from 16 channels to 8 channels U3
        result_conv_13 = self.u3(act_2_fine)  # +result_conv_2_fined
        normed_batch_13 = self.final_norm(result_conv_13)
        result_prelu_13 = normed_batch_13  # s_sigmoid(normed_batch_13)  # tf.nn.leaky_relu(normed_batch_13,name='relu_1'）

        # averge over the remaining channels as the output

        prediction = tf.reduce_mean(result_prelu_13, axis=4,
                                    keepdims=True)  # /tf.reduce_max(result_prelu_14)#*tf.reduce_max(normed_batch) #tf.multiply(result_prelu_7,result_prelu_1)

        return (prediction, estimatation1)
