﻿# 5#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 19 17:10:32 2018

@author: lee
with batch normalization
"""
import datetime




import os
import logging
import time
######## configuration variables #######

base_dir = "/data"
ground_truthdir = base_dir + '/train/ground_truth/'
input_dir = base_dir + "/train/input/"

test_image = base_dir + "/test/input/19.npz"  #image to use for the test output every 300 steps. Set None to disable this behaviour

learning_rate = 0.008 # Learning Rate
model_type = "full" # full to use the model as in paper "reduced" to use the model without the 'update' step

input_crop_shape = (1000,1000,1) #set to None for whole image
number_of_images_to_load =200 #set to -1 for all images
epochs = 400

######## Setup Folders #######
timestr = time.strftime("%Y%m%d-%H%M%S")
run_name = "/RLN3D_Normed_Model_Clipped_{}_{}_{}/".format(model_type,learning_rate,timestr)

train_model_path = base_dir + '/train/model_rl' + run_name
train_output = base_dir + '/train/output_rl' + run_name
test_output = base_dir + '/test/output_rl' + run_name
log_dir = base_dir + "/logs" + run_name

if not os.path.exists(train_model_path):
    os.makedirs(train_model_path)

if not os.path.exists(train_output):
    os.makedirs(train_output)

if not os.path.exists(test_output):
    os.makedirs(test_output)

if not os.path.exists(log_dir):
    os.makedirs(log_dir)

logging.basicConfig(filename=log_dir+"out.log",filemode="a",encoding='utf-8', level=logging.DEBUG)

import TF2ImageHelpers as images #imported here to  redirect logging
import tensorflow as tf
from RLN_single_Model import RLN_model, RLN_model_simple






#### helpers for the loss function ####
@tf.function
def get_mae(y_true, y_pred):
    return tf.reduce_mean(tf.abs(y_true - y_pred))

@tf.function
def get_mse(y_true, y_pred):
    return tf.reduce_mean(tf.square(y_true - y_pred)) + 0.0001

@tf.function
def get_max_se(y_true, y_pred):
    return tf.reduce_max(tf.square(y_true - y_pred ))
@tf.function
def get_SSIM(self, gt_label, dl_op, max_val=1):
    mean_prediction = tf.reduce_mean(dl_op)
    mean_gt = tf.reduce_mean(gt_label)
    sigma_prediction = tf.reduce_mean(tf.square(tf.subtract(dl_op, mean_prediction)))
    sigma_gt = tf.reduce_mean(tf.square(tf.subtract(gt_label, mean_gt)))
    sigma_cross = tf.reduce_mean(tf.multiply(tf.subtract(dl_op, mean_prediction),
                                             tf.subtract(gt_label, mean_gt)))
    SSIM_1 = 2 * tf.multiply(mean_prediction, mean_gt) + 1e-4 * max_val * max_val
    SSIM_2 = 2 * sigma_cross + 9e-4 ** max_val * max_val
    SSIM_3 = tf.square(mean_prediction) + tf.square(mean_gt) + 1e-4 ** max_val * max_val
    SSIM_4 = sigma_prediction + sigma_gt + 9e-4 ** max_val * max_val
    SSIM = tf.div(tf.multiply(SSIM_1, SSIM_2), tf.multiply(SSIM_3, SSIM_4))
    return SSIM


def get_correlation(x,y):
    u_x = tf.reduce_mean(x)
    u_y = tf.reduce_mean(y)
    sigma_xy = tf.reduce_mean((x-u_x)*(y-u_y))
    sigma_x = tf.math.reduce_std(x)
    sigma_y = tf.math.reduce_std(y)

    return (sigma_xy + 0.002) / (sigma_x*sigma_y+0.002)

def loss(y_true, prediction, estimatation1, update1,step):

    mse = 1.0 * get_mse(prediction, y_true) #+ 0.5 * get_max_se(prediction, y_true)
    tf.summary.scalar("squared_error", mse,step)
    SSIM = get_SSIM(y_true, prediction)
    tf.summary.scalar("ssim", SSIM,step)
    estimation_mse = get_mse(estimatation1, y_true)
    tf.summary.scalar("estimation_mse", estimation_mse,step)
    update_mse = get_mse(y_true, update1)
    tf.summary.scalar("update_mse", update_mse,step)

    #mse2 = estimation_mse  # -tf.log((1+self.SSIM2)/2)#-tf.log((1+self.SSIM1)/2)
    # self.loss =0.1*self.mse2+1*self.mse-1.0*tf.log((1+self.SSIM)/2)#-k1*self.prediction_min

    loss = 1 * mse - 1 * tf.math.log((1 + SSIM) / 2) + 0.1 * estimation_mse + 0.1 * update_mse
    tf.summary.scalar("loss", loss, step)
    return loss


def loss_simple(y_true, prediction, estimatation1, step):

    mse = 1.0 * get_mse(prediction, y_true) #+ 0.5 * get_max_se(prediction, y_true)
    tf.summary.scalar("squared_error", mse,step)
    SSIM = get_SSIM(y_true, prediction)
    tf.summary.scalar("ssim", SSIM,step)
    estimation_mse = get_mse(estimatation1, y_true)
    tf.summary.scalar("estimation_mse", estimation_mse,step)


    #mse2 = estimation_mse  # -tf.log((1+self.SSIM2)/2)#-tf.log((1+self.SSIM1)/2)
    # self.loss =0.1*self.mse2+1*self.mse-1.0*tf.log((1+self.SSIM)/2)#-k1*self.prediction_min

    loss = 1 * mse - 1 * tf.math.log((1 + SSIM) / 2) + 0.1 * estimation_mse
    tf.summary.scalar("loss", loss, step)
    return loss


#### Training Setup ####

model = None

if model_type =="full":
    model = RLN_model(name="test")
if model_type == "reduced":
    model = RLN_model_simple(name="test")

with tf.device('/CPU:0'): #load dataset to CPU memory
    dataset_original, (mean,var) = images.move_dataset_to_mem(input_dir, ground_truthdir, number_of_images_to_load)

optimizer = tf.keras.optimizers.Adam(learning_rate=learning_rate)

train_summary_writer = tf.summary.create_file_writer(log_dir)

i = tf.Variable(0, trainable=False, dtype=tf.int64)

ckpt = tf.train.Checkpoint(step=i, model=model)
manager = tf.train.CheckpointManager(ckpt, train_model_path, max_to_keep=20)

#### training Loop ####

for epoch in range(0, epochs):
    #for each epoch shuffle the dataset
    logging.info("\nStart of epoch %d" % (epoch,))
    dataset = dataset_original.map(lambda x, y: images.randomize(input_crop_shape, x, y))
    dataset = dataset.shuffle(buffer_size=10).batch(batch_size=16)
    logging.debug("Memory Usage:  GPU: {}   CPU: {}".format(tf.config.experimental.get_memory_info('GPU:0'),tf.config.experimental.list_physical_devices('CPU:0')))


    for (x_batch_train, y_batch_train) in dataset: # loop over the batches
        std = tf.math.reduce_std(x_batch_train, axis=(1, 2, 3, 4), keepdims=True)
        mean = tf.math.reduce_mean(x_batch_train, axis=(1, 2, 3, 4), keepdims=True)

        normed = (x_batch_train - mean) / std
        #logging.debug("input std: {}, mean: {} ; normed std: {}, mean: {}".format(std.numpy(),mean.numpy(),tf.math.reduce_std(normed,axis=(1,2,3)),tf.math.reduce_std(normed,axis=(1,2,3))) )
        # logging.debug(f"input dtype : {x_batch_train.dtype}, norm dtype: {normed.dtype}, norm shape:{normed.shape}, mean:{mean.shape}, std:{std.shape}, {std.dtype}")
        with train_summary_writer.as_default(), tf.GradientTape() as tape: # training step

            prediction,estimatation1,update1,loss_value = None,None,None,None

            if model_type == "full":
                prediction, estimatation1, update1 = model(normed, training=True)
                prediction = prediction*std + mean
                estimatation1 = estimatation1*std + mean
                update1 = update1*std + mean

                prediction = tf.clip_by_value(prediction,0,1,name="final_activation_clip")
                estimatation1 = tf.clip_by_value(estimatation1,0,1,name="final_activation_clip")
                update1 = tf.clip_by_value(update1,0,1,name="final_activation_clip")

                loss_value = loss(y_batch_train, prediction, estimatation1, update1, i)

            if model_type == "reduced":
                prediction, estimatation1 = model(x_batch_train, training=True)
                loss_value = loss_simple(y_batch_train, prediction, estimatation1, i)

        grads = tape.gradient(loss_value, model.trainable_weights)
        optimizer.apply_gradients(zip(grads, model.trainable_weights))

        i.assign_add(1)

        with train_summary_writer.as_default():
            if int(i) % 50 == 0: #log images every 50 iters
                tf.summary.image("input", x_batch_train, i,max_outputs=3)
                tf.summary.image("ground_truth", y_batch_train, i,max_outputs=3)
                tf.summary.image("output", prediction, i,max_outputs=3)
                tf.summary.image("estimation", estimatation1, i,max_outputs=3)
                if model_type == "full" and update1 is not None:
                    tf.summary.image("update", update1, i,max_outputs=3)

            if int(i) % 200 == 0: # drop a checkpoint every 200  iters
                save_path = manager.save()
                logging.info("Saved checkpoint for step {}: {}".format(int(ckpt.step), save_path))
                logging.info("loss {:1.2f}".format(loss_value.numpy()))

            if int(i)%301 == 0 and test_image is not None: # run a full inference every 301 iters
                logging.info("Running inference")
                x = tf.expand_dims(images.read_file(test_image), axis=0) #prepare the array
                tf.summary.image("test input", x, i,max_outputs=1)

                std = tf.math.reduce_std(x, axis=(1, 2, 3, 4), keepdims=True) #normalize the input
                mean = tf.math.reduce_mean(x, axis=(1, 2, 3, 4), keepdims=True)
                x = (x - mean) / std

                prediction, estimatation1, update1 = model(x, training=False) #denormalize the input

                prediction = prediction*std + mean
                estimatation1 = estimatation1*std + mean
                update1 = update1*std + mean

                prediction = tf.clip_by_value(prediction,0,1,name="final_activation_clip")
                estimatation1 = tf.clip_by_value(estimatation1,0,1,name="final_activation_clip")
                update1 = tf.clip_by_value(update1,0,1,name="final_activation_clip")


                tf.summary.image("test output", prediction, i,max_outputs=1)
                tf.summary.image("test estimation", estimatation1, i,max_outputs=1)

        logging.info("Step %s", int(i))
