import os
import sys
import time

import numpy as np
import tensorflow as tf

import TF2ImageHelpers as images
from RLN_single_Model import RLN_model_simple, RLN_model

base_dir = "/data"
ground_truthdir = base_dir + '/train/ground_truth/'
input_dir = base_dir + "/test/input/"

timestr = time.strftime("%Y%m%d-%H%M%S")
run_name = "/{}/".format(sys.argv[1])

train_model_path = base_dir + '/train/model_rl' + run_name
train_output = base_dir + '/train/output_rl' + run_name
test_output = base_dir + '/test/output_rl' + run_name
log_dir = base_dir + "/logs" + run_name
print(test_output)

if not os.path.exists(train_model_path) or not os.path.exists(train_output) or not os.path.exists(
        test_output) or not os.path.exists(log_dir):
    raise Exception("missing locations")

import logging

logging.basicConfig(filemode="a", encoding='utf-8', level=logging.DEBUG)



model = RLN_model(name="test")

i = tf.Variable(0, trainable=False, dtype=tf.int64)

ckpt = tf.train.Checkpoint(step=i, model=model)
manager = tf.train.CheckpointManager(ckpt, train_model_path, max_to_keep=20)
ckpt.restore(manager.latest_checkpoint)
logging.info("restore sucessfull")
a = os.listdir(input_dir)
a = [i for i in a if i.split('.')[-1] == 'npz']
a = sorted(a, key=lambda x: int(x.split(".")[0]))
logging.debug(a)
for path_i in a:
    logging.debug(path_i)
    x = images.read_file(input_dir + path_i)
    x = tf.expand_dims(x, 0)

    st = time.time()

    std = tf.math.reduce_std(x, axis=(1, 2, 3, 4), keepdims=True)
    mean = tf.math.reduce_mean(x, axis=(1, 2, 3, 4), keepdims=True)

    x = (x - mean) / std
    x_i = model(x, training=False)
    x_i = x_i[0] * std + mean
    x_i = tf.clip_by_value(x_i, 0, 1, name="final_activation_clip")

    et = time.time()
    dt = np.array(et - st)
    np.savez_compressed(test_output + path_i + ".npz", x=x, f_x=x_i, dt=dt)
    et2 = time.time()
    i.assign_add(1)
    logging.info("Run: {} calc time: {} save time: {} memory: {}".format(i.numpy(), dt, et2 - et,
                                                                         tf.config.experimental.get_memory_info(
                                                                             'GPU:0')))
