import pickle

from helpers.image import show_image,read_y10
from helpers.output import save_array,load_array
import os
import cv2
import numpy as np
import matplotlib.pyplot as plt

"""
this script is for actually determining the PSF for each measurement. For each measurement it will do:

1) It finds the brightest point in the image and cuts a square with (2*delta+1) side length around the point.
2) It will show the whole image and teh cut portion of the image. You can accept by [ENTER] or press c to cancel
3) It will show the normed version of the cropped image (L1 norm). You can accept by [ENTER] or press c to cancel
4) It will save the image
"""

delta = 3# box size
detail_display = False # display the position of the brightest point on the whole image
root = ("/home/guney/Project/Richardson-Lucy-Net/helper scripts/data/psf2")
dirs = os.listdir(root)
for i,path in enumerate(dirs): # iterate over the measurements
    path = root + "/" + path+"/"

    print(path)
    try:
        image = load_array(path+"avg_image.npz")
    except FileNotFoundError:
        continue

    max_i = np.unravel_index(np.argmax(image),image.shape) # find the brightest point

    print("Maximal Brightness at {}".format(max_i))

    y_min, y_max = max_i[0]-delta,max_i[0]+delta+1 #determine the box around the point
    x_min, x_max = max_i[1]-delta,max_i[1]+delta+1

    top_left_point = (x_min,y_min)
    bottom_right_point = (x_max,y_max)
    if detail_display:
        im_2 = np.copy(image)
        cv2.rectangle(im_2,top_left_point,bottom_right_point,1.0,4)
        show_image(im_2,"avg",norm=(0,1))
    cropped = image[y_min:y_max, x_min:x_max]
    #print(np.unravel_index(np.argmax(cropped),cropped.shape))
    show_image(cropped,"avg_cropped",norm=(0,4))
    #print(cropped.shape)
    if cv2.waitKey() == ord('c'):
        print("exiting")
        cv2.destroyAllWindows()
        exit(0)

    avg_image = cropped


    normed = np.zeros(np.shape(avg_image), dtype=np.float64)
    cv2.normalize(avg_image,normed,1.0,0,cv2.NORM_L1) # normalize to L! norm

    show_image(normed,"avg_L1",norm=(0,3))

    plt.imshow(normed)
    plt.savefig(path+"psf.png")
    plt.show()

    print("L1 norm is {}".format(np.sum(normed)))
    if cv2.waitKey() == ord('c'):
        print("exiting")
        cv2.destroyAllWindows()
        exit(0)

    plt.figure()
    plt.imshow(normed)
    plt.savefig(path+"psf.png")
    plt.show()

    crop_limits = {"y_max" : y_max, "y_min" : y_min, "x_max" : x_max, "x_min" :x_min}
    with open(path+'crop_limits.pickle', 'wb') as handle:
        pickle.dump(crop_limits, handle, protocol=pickle.HIGHEST_PROTOCOL)
    #cv2.imwrite(path+"psf.png",normed)
    np.save(path+"cropped.npy",cropped)
    np.save(path+"psf.npy",normed)

    cv2.destroyAllWindows()