import os

import cv2
import matplotlib.pyplot as plt
import numpy as np

from helpers.image import show_image

path ="/home/guney/Project/Richardson-Lucy-Net/pic_proc/data/output_from_server/"
a = os.listdir(path)
a = [path+i for i in a if i.split('.')[-1] == 'npz']

display = False
mse_f_x = []
mse_x = []

max_se_f_x = []
max_se_x =[]

var_sim_f_x =[]
var_sim_x_ = []

"""
This script is used to display the output of the test that runs on the server. The script 
RLN_single_Test.py outputs .npz files with the arrays "x" for input, "y" for ground truth, and "f_x" for 
inference output. 
This displays 
- gt and x, f_x for all .npz s in the path.
- the diff between gt and f_x is calculated and shown where red means pixel value smaller in f_x and green the opposite.
- some other graphs
"""
def var_similarity_multiscale(x,y,s):

    sum = []
    for i in range(0,x.shape[0],x.shape[0]//s):
        for j in range(0,x.shape[1],x.shape[1]//s):
            x_i = x[i:i+x.shape[0]//s,j:j+x.shape[1]//s]
            y_i = y[i:i+x.shape[0]//s,j:j+x.shape[1]//s]
            sum_i = var_similarity(x_i,y_i)
            if sum_i is not None:
                sum.append(sum_i)
    return np.mean(sum)
def var_similarity(x,y):
    cov_m = np.cov(x.flatten(),y.flatten())
    if cov_m[0,0] == 0 or cov_m[1,1] == 0:
        return None

    return cov_m[0,1]/np.sqrt(cov_m[0,0]*cov_m[1,1])

def squared_error(x,y):
    delta = (x.flatten() - y.flatten())**2
    return delta

for data in a[:50]:
    print("==========================================================")
    print(data)
    print("==========================================================")

    data = np.load(data,allow_pickle=True)
    inference = data["f_x"][0]
    input = data["x"][0]
    gt = data["y"][0]
    if display:
        show_image(inference,"inference")
        show_image(gt,"gt")
        show_image(input,"input")

        # Compute the absolute difference between the images
        diff = inference - gt
        diff = np.where(diff<0,0.,diff)
        _, thresholded_diff = cv2.threshold(diff, 0.1, 1.0, cv2.THRESH_BINARY)
        thresholded_diff_bgr = cv2.cvtColor(thresholded_diff, cv2.COLOR_GRAY2BGR)
        thresholded_diff_bgr[np.where((thresholded_diff_bgr == [1.0, 1.0, 1.0]).all(axis=2))] = [0, 0, 1.0]

        diff = gt - inference
        diff = np.where(diff<0,0.,diff)
        _, thresholded_diff = cv2.threshold(diff, 0.1, 1.0, cv2.THRESH_BINARY)
        thresholded_diff_bgr_2 = cv2.cvtColor(thresholded_diff, cv2.COLOR_GRAY2BGR)
        thresholded_diff_bgr[np.where((thresholded_diff_bgr_2 == [1.0, 1.0, 1.0]).all(axis=2))] += [0, 1.0, 0.0]

        highlighted_image = cv2.addWeighted(cv2.cvtColor(inference, cv2.COLOR_GRAY2BGR), 0.3, thresholded_diff_bgr, 0.7, 0)
        show_image(highlighted_image,"diff")


        if cv2.waitKey() == ord('c'):
            print("exiting")
            cv2.destroyAllWindows()
            exit(0)
    print()
    for s in range(1,2):
         print("scale: {} s(gt,f(x)) = {}, s(gt,x) = {}".format(s,var_similarity_multiscale(gt,inference,s),var_similarity_multiscale(gt,input,s)))

    print("==========================================================")


    delta1 = squared_error(gt,inference)
    delta2 = squared_error(gt,input)
    # fig, ax = plt.subplots(1, 2, figsize=(15, 10))
    # ax[0].hist(delta1, bins="auto")
    # ax[0].set_yscale("log")
    #
    # ax[1].hist(delta2, bins="auto")
    # ax[1].set_yscale("log")
    # fig.show()
    mse_f_x.append(np.mean(delta1))
    mse_x.append(np.mean(delta2))

    max_se_f_x.append(np.max(delta1))
    max_se_x.append(np.max(delta2))

    var_sim_x_.append(var_similarity(gt,input))
    var_sim_f_x.append(var_similarity(gt,inference))



    # print("mse(gt,f_x) = {}, mse(gt,x) = {}, max_se(gt,f_x) = {}, max_se(gt,x) = {}".format(np.mean(delta1),np.mean(delta2),np.max(delta1),np.max(delta2)))


fig, ax = plt.subplots(2, 3,figsize=(30,20))
ax[0,0].hist(mse_f_x, bins="auto")
ax[0,0].set_title("MSE f(x)")

ax[1,0].hist(mse_x, bins="auto")
ax[1,0].set_title("MSE x")

ax[0,1].hist(max_se_f_x, bins="auto")
ax[0,1].set_title("MAX SE f(x)")

ax[1,1].hist(max_se_x, bins="auto")
ax[1,1].set_title("MAX SE x")

ax[0,2].hist(var_sim_f_x, bins="auto")
ax[0,2].set_title("Similarity f(x)")

ax[1,2].hist(var_sim_x_, bins="auto")
ax[1,2].set_title("Similarity x")
fig.show()