import numpy as np
import cv2


def show_image(image,title,width=800,height=600,norm=None):
    """
    shows image using an OpenCV window of width and height. Uses title as the window name.
    :param image_in: Numpy Array as OpenCV image
    :param title: used for the window name
    :param width: window width
    :param height: window height
    """
    disp = np.zeros_like(image)
    if norm is not None:
        cv2.normalize(image,disp,norm[0],norm[1],cv2.NORM_MINMAX)
    else:
        disp = image

    cv2.namedWindow(title, cv2.WINDOW_NORMAL)
    cv2.resizeWindow(title,width, height)
    cv2.imshow(title, disp)
def read_y10(filepath,width=3840, height = 2192):
    """
    Reads Y10 file and returns Numpy Array as OpenCV image.
    :param width: Used to reshape the 1D Array to an image
    :param height: Same as width
    :param filepath:
    :return:OpenCv Image
    :raise Dimension Mismatch: if the number of 16 bit words don't match the pixel count
    """
    data = np.fromfile(filepath, dtype=np.uint16)
    if height * width != len(data):
        raise Exception("Dimension Mismatch at {}".format(filepath))
    #data = data * 2**6 The Data is multiplied by 2^6 to shift the bytes to left.
    return np.reshape(data,(height,width))

def display_y10(filepath,name=None,normalize=True):
    """
    display an Y10 image using OpenCV window. Uses for width and height the defaults for the sony CCD.

    :param filepath: Filepath of the Y10
    :param name: display image if None uses filepath as the name
    """
    data = read_y10(filepath,3840,2192)
    image = np.zeros(np.shape(data),dtype=np.uint16)
    if normalize:
        cv2.normalize(data,image,0,65535,cv2.NORM_MINMAX)
    else:
        image = data

    if name is None:
        name = filepath
    show_image(image,name)


def display_and_wait_y10(filepath, name=None, normalize=True):
    """
    display an Y10 image using OpenCV window. Uses for width and height the defaults for the sony CCD.

    :param filepath: Filepath of the Y10
    :param name: display image if None uses filepath as the name
    """
    data = read_y10(filepath, 3840, 2192)
    image = np.zeros(np.shape(data), dtype=np.uint16)
    if normalize:
        cv2.normalize(data, image, 0, 65535, cv2.NORM_MINMAX)
    else:
        image = data

    if name is None:
        name = filepath
    show_image(image, name)
    cv2.waitKey(0)
    cv2.destroyWindow(name)






