import os

import numpy as np


def get_last_run_number(path):
    # Get a list of all existing output folders
    existing_folders = [folder for folder in os.listdir(path) if folder.startswith("output_")]
    # If there are no existing output folders, return 0
    if not existing_folders:
        return 0

    # Extract run numbers from folder names and find the maximum
    run_numbers = [int(folder.split("_")[1]) for folder in existing_folders]
    last_run_number = max(run_numbers)

    return last_run_number

def get_save_paths(output_path,with_run_path = False):
    """
    Get the paths $output_path$/output_N/gt for the ground truth and $output_path$/output_N/blurred for saving images to.
    :param output_path: base output location
    :return: ground truth save location, blurred save location
    """
    last_run = get_last_run_number(output_path)
    run_path = output_path + f"output_{last_run + 1}/"
    os.makedirs(run_path)
    gt_path = run_path + "ground_truth/"
    blur_path = run_path + "input/"

    os.makedirs(gt_path)
    os.makedirs(blur_path)
    if with_run_path:
        return gt_path,blur_path,run_path
    return gt_path, blur_path

def save_array(path,array):
    np.savez_compressed(path,array)

def load_array(path):
    a = np.load(path)
    b = a[a.files[0]]
    return b